package de.mpg.mpiwg.berlin.mpdl.lt.analyzer;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;

import de.mpg.mpiwg.berlin.mpdl.exception.ApplicationException;
import de.mpg.mpiwg.berlin.mpdl.lt.doc.regularization.Regularization;
import de.mpg.mpiwg.berlin.mpdl.lt.doc.regularization.RegularizationManager;
import de.mpg.mpiwg.berlin.mpdl.lt.general.Language;

public class MpdlNormalizer {
  static final private String IT_VOWELS = "AEIOUaeiou" +
                                          "\u00c6\u00e6" + // AE ligatures
                                          "\u0152\u0153";  // OE ligatures
  static final private String IT_CONS = "BCDFGHKLMNPQRSTVWXZ" +
                                        "bcdfghklmnpqrstvwxz" +
                                        "\u017f\u00df";  // long/sharp S
  private String[] normFunctionsToUse = {"reg", "norm"};  // default is to use all of these normalization functions
  private String language;
  private int[] offsets;
  
  public MpdlNormalizer(String[] normFunctionsToUse, String lang) {
    this.normFunctionsToUse = normFunctionsToUse;
    String language = Language.getInstance().getLanguageId(lang); 
    this.language = language;
  }

  public MpdlNormalizer(String language) {
    this.language = language;
  }

  /**
   * Applies the normalization rules in <code>language</code> to
   * <code>s</code>, without offset tracking.
   *
   * @param s            source string
   * @return             normalized string
   */
  public String normalize(String s) throws ApplicationException {
    String normStr = s;
    if (useRegFunction()) {
      // try to regularize the string to the norm form over predefined regularizations
      RegularizationManager regManager = RegularizationManager.getInstance();
      ArrayList<Regularization> regs = regManager.findRegsByOrig(language, s);
      if (regs != null && regs.size() > 0) {
        Regularization reg = regs.get(0);  // only one: the first one
        String regNormStr = reg.getNorm();
        normStr = regNormStr;
      }
    }
    if (useNormFunction()) {
      // normalize the string by string replace
      normStr = normalize(normStr, null);
    }
    return normStr;
  }

  private boolean useRegFunction() {
    boolean useReg = false;
    for (int i=0; i< normFunctionsToUse.length; i++) {
      String function = normFunctionsToUse[i];
      if (function.equals("reg"))
        return true;
    }
    return useReg;
  }
  
  private boolean useNormFunction() {
    boolean useNorm = false;
    for (int i=0; i< normFunctionsToUse.length; i++) {
      String function = normFunctionsToUse[i];
      if (function.equals("norm"))
        return true;
    }
    return useNorm;
  }

  /**
   * Applies the normalization rules in <code>language</code> to
   * <code>s</code>, with offset tracking.<p>
   *
   * <strong>WARNING:</strong>
   * Arboreal will not work properly if a normalization substitution
   * replaces a source character with more than two target characters!
   * This is simply a BUG, and should be fixed. Fortunately, however,
   * one does not often need such a replacement.<p>
   *
   * @param s            source string
   * @param offsets      character offset table
   * @return             normalized string
   */
  public String normalize(String s, int[] offsets) {
    this.offsets = offsets;
    if (language.equals("la") || language.equals("lat")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case 'j': replace = "i"; break;
          case 'v': replace = "u"; break;
          /*
           * Linguistic note: /u/ and /v/ are rarely phonemic
           * in Latin, as in alui 's/he nourished' vs.
           * alvi 'of a belly', volui 's/he wished' or 'it rolled'
           * vs. volvi 'to be rolled', (in)seruit 's/he joined
           * together' vs. (in)servit 's/he serves'.
           */
          case 'q':
            if ((i < s.length() - 1) && (s.charAt(i + 1) == ';'))
              replace = "qu";
            else
              replace = "q";
            break;
          case ';':
            if ((i > 0) && (s.charAt(i - 1) == 'q'))
              replace = "e";
            else if ((i == 0) || ! Character.isLetter(s.charAt(i - 1)))
              replace = ";";
            else
              replace = "";
            break;
          case '\u0300': replace = ""; break; // COMBINING GRAVE ACCENT
          case '\u0301': replace = ""; break; // COMBINING ACCUTE ACCENT
          case '\u0302': replace = ""; break; // COMBINING CIRCUMFLEX ACCENT
    
          case '\u00c0': replace = "A"; break; // LATIN CAPITAL LETTER A GRAVE
          case '\u00c1': replace = "A"; break; // LATIN CAPITAL LETTER A ACUTE
          case '\u00c2': replace = "A"; break; // LATIN CAPITAL LETTER A CIRCUMFLEX
          case '\u00c4': replace = "A"; break; // LATIN CAPITAL LETTER A DIAERESIS
          case '\u00c6': replace = "Ae"; break; // LATIN CAPITAL LETTER A E
          case '\u00c7': replace = "C"; break; // LATIN CAPITAL LETTER C CEDILLA
          case '\u00c8': replace = "E"; break; // LATIN CAPITAL LETTER E GRAVE
          case '\u00c9': replace = "E"; break; // LATIN CAPITAL LETTER E ACUTE
          case '\u00ca': replace = "E"; break; // LATIN CAPITAL LETTER E CIRCUMFLEX
          case '\u00cb': replace = "E"; break; // LATIN CAPITAL LETTER E DIAERESIS   
          case '\u00cc': replace = "I"; break; // LATIN CAPITAL LETTER I GRAVE;
          case '\u00cd': replace = "I"; break; // LATIN CAPITAL LETTER I ACUTE
          case '\u00ce': replace = "I"; break; // LATIN CAPITAL LETTER I CIRCUMFLEX
          case '\u00cf': replace = "I"; break; // LATIN CAPITAL LETTER I DIAERESIS
          case '\u00d2': replace = "O"; break; // LATIN CAPITAL LETTER O GRAVE
          case '\u00d3': replace = "O"; break; // LATIN CAPITAL LETTER O ACUTE
          case '\u00d4': replace = "O"; break; // LATIN CAPITAL LETTER O CIRCUMFLEX
          case '\u00d6': replace = "O"; break; // LATIN CAPITAL LETTER O DIAERESIS
          case '\u00d9': replace = "U"; break; // LATIN CAPITAL LETTER U GRAVE
          case '\u00da': replace = "U"; break; // LATIN CAPITAL LETTER U ACUTE
          case '\u00db': replace = "U"; break; // LATIN CAPITAL LETTER U CIRCUMFLEX
          case '\u00dc': replace = "U"; break; // LATIN CAPITAL LETTER U DIAERESIS 
          case '\u00e0': replace = "a"; break; // LATIN SMALL LETTER A GRAVE
          case '\u00e1': replace = "a"; break; // LATIN SMALL LETTER A ACUTE
          case '\u00e2': replace = "a"; break; // LATIN SMALL LETTER A CIRCUMFLEX
          case '\u00e4': replace = "a"; break; // LATIN SMALL LETTER A DIAERESIS
          case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
          case '\u00e7': replace = "c"; break; // LATIN SMALL LETTER C CEDILLA
          case '\u00e8': replace = "e"; break; // LATIN SMALL LETTER E GRAVE
          case '\u00e9': replace = "e"; break; // LATIN SMALL LETTER E ACUTE
          case '\u00ea': replace = "e"; break; // LATIN SMALL LETTER E CIRCUMFLEX
          case '\u00eb': replace = "e"; break; // LATIN SMALL LETTER E DIAERESIS
          case '\u00ec': replace = "i"; break; // LATIN SMALL LETTER I GRAVE
          case '\u00ed': replace = "i"; break; // LATIN SMALL LETTER I ACUTE
          case '\u00ee': replace = "i"; break; // LATIN SMALL LETTER I CIRCUMFLEX
          case '\u00ef': replace = "i"; break; // LATIN SMALL LETTER I DIAERESIS
          case '\u00f2': replace = "o"; break; // LATIN SMALL LETTER O GRAVE
          case '\u00f3': replace = "o"; break; // LATIN SMALL LETTER O ACUTE
          case '\u00f4': replace = "o"; break; // LATIN SMALL LETTER O CIRCUMFLEX
          case '\u00f6': replace = "o"; break; // LATIN SMALL LETTER O DIAERESIS
          case '\u00f9': replace = "u"; break; // LATIN SMALL LETTER U GRAVE
          case '\u00fa': replace = "u"; break; // LATIN SMALL LETTER U ACUTE
          case '\u00fb': replace = "u"; break; // LATIN SMALL LETTER U CIRCUMFLEX
          case '\u00fc': replace = "u"; break; // LATIN SMALL LETTER U DIAERESIS
          case '\u0100': replace = "A"; break; // LATIN CAPITAL LETTER A MACRON
          case '\u0101': replace = "a"; break; // LATIN SMALL LETTER A MACRON
          case '\u0102': replace = "A"; break; // LATIN CAPITAL LETTER A BREVE
          case '\u0103': replace = "a"; break; // LATIN SMALL LETTER A BREVE
          case '\u0112': replace = "E"; break; // LATIN CAPITAL LETTER E MACRON
          case '\u0113': replace = "e"; break; // LATIN SMALL LETTER E MACRON
          case '\u0114': replace = "E"; break; // LATIN CAPITAL LETTER E BREVE
          case '\u0115': replace = "e"; break; // LATIN SMALL LETTER E BREVE
          case '\u0118': replace = "Ae"; break; // LATIN CAPITAL LETTER E OGONEK
          case '\u0119': replace = "ae"; break; // LATIN SMALL LETTER E OGONEK
          case '\u012a': replace = "I"; break; // LATIN CAPITAL LETTER I MACRON
          case '\u012b': replace = "i"; break; // LATIN SMALL LETTER I MACRON
          case '\u012c': replace = "I"; break; // LATIN CAPITAL LETTER I BREVE
          case '\u012d': replace = "i"; break; // LATIN SMALL LETTER I BREVE
          case '\u014c': replace = "O"; break; // LATIN CAPITAL LETTER O MACRON
          case '\u014d': replace = "o"; break; // LATIN SMALL LETTER O MACRON
          case '\u014e': replace = "O"; break; // LATIN CAPITAL LETTER O BREVE
          case '\u014f': replace = "o"; break; // LATIN SMALL LETTER O BREVE
          case '\u0152': replace = "Oe"; break; // LATIN CAPITAL LETTER O E
          case '\u0153': replace = "oe"; break; // LATIN SMALL LETTER O E
          case '\u016a': replace = "U"; break; // LATIN CAPITAL LETTER U MACRON
          case '\u016b': replace = "u"; break; // LATIN SMALL LETTER U MACRON
          case '\u016c': replace = "U"; break; // LATIN CAPITAL LETTER U BREVE
          case '\u016d': replace = "u"; break; // LATIN SMALL LETTER U BREVE
          case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
          case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
          case '\u00ad': break; // soft hyphen
          // new in MPDL project by J. Willenborg
          case '\u1e14': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e15': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e16': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e17': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e18': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e19': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e1a': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e1b': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e1c': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1e1d': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1eb8': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1eb9': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1eba': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ebb': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ebc': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ebd': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ebe': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ebf': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec0': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec1': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec2': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec3': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec4': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec5': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec6': replace = "e"; break; // LATIN ... LETTER E WITH ...
          case '\u1ec7': replace = "e"; break; // LATIN ... LETTER E WITH ...
          // by Malcolm
          case '\u2329': break; // BRA
          case '\u232a': break; // KET
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2) 
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("it")) {
      // new Mpdl code: added by J. Willenborg: some of Malcolms code did not work without errors so it has to be taken away, also all latin stuff is imported
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u00c0': replace = "A"; break; // LATIN CAPITAL LETTER A GRAVE
          case '\u00c1': replace = "A"; break; // LATIN CAPITAL LETTER A ACUTE
          case '\u00c2': replace = "A"; break; // LATIN CAPITAL LETTER A CIRCUMFLEX
          case '\u00c4': replace = "A"; break; // LATIN CAPITAL LETTER A DIAERESIS
          case '\u00c6': replace = "Ae"; break; // LATIN CAPITAL LETTER A E
          case '\u00c7': replace = "C"; break; // LATIN CAPITAL LETTER C CEDILLA
          case '\u00c8': replace = "E"; break; // LATIN CAPITAL LETTER E GRAVE
          case '\u00c9': replace = "E"; break; // LATIN CAPITAL LETTER E ACUTE
          case '\u00ca': replace = "E"; break; // LATIN CAPITAL LETTER E CIRCUMFLEX
          case '\u00cb': replace = "E"; break; // LATIN CAPITAL LETTER E DIAERESIS
          case '\u00cc': replace = "I"; break; // LATIN CAPITAL LETTER I GRAVE;
          case '\u00cd': replace = "I"; break; // LATIN CAPITAL LETTER I ACUTE
          case '\u00ce': replace = "I"; break; // LATIN CAPITAL LETTER I CIRCUMFLEX
          case '\u00cf': replace = "I"; break; // LATIN CAPITAL LETTER I DIAERESIS
          case '\u00d2': replace = "O"; break; // LATIN CAPITAL LETTER O GRAVE
          case '\u00d3': replace = "O"; break; // LATIN CAPITAL LETTER O ACUTE
          case '\u00d4': replace = "O"; break; // LATIN CAPITAL LETTER O CIRCUMFLEX
          case '\u00d6': replace = "O"; break; // LATIN CAPITAL LETTER O DIAERESIS
          case '\u00d9': replace = "U"; break; // LATIN CAPITAL LETTER U GRAVE
          case '\u00da': replace = "U"; break; // LATIN CAPITAL LETTER U ACUTE
          case '\u00db': replace = "U"; break; // LATIN CAPITAL LETTER U CIRCUMFLEX
          case '\u00dc': replace = "U"; break; // LATIN CAPITAL LETTER U DIAERESIS
          case '\u00e0': replace = "a"; break; // LATIN SMALL LETTER A GRAVE
          case '\u00e1': replace = "a"; break; // LATIN SMALL LETTER A ACUTE
          case '\u00e2': replace = "a"; break; // LATIN SMALL LETTER A CIRCUMFLEX
          case '\u00e4': replace = "a"; break; // LATIN SMALL LETTER A DIAERESIS
          case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
          case '\u00e7': replace = "c"; break; // LATIN SMALL LETTER C CEDILLA
          case '\u00e8': replace = "e"; break; // LATIN SMALL LETTER E GRAVE
          case '\u00e9': replace = "e"; break; // LATIN SMALL LETTER E ACUTE
          case '\u00ea': replace = "e"; break; // LATIN SMALL LETTER E CIRCUMFLEX
          case '\u00eb': replace = "e"; break; // LATIN SMALL LETTER E DIAERESIS
          case '\u00ec': replace = "i"; break; // LATIN SMALL LETTER I GRAVE
          case '\u00ed': replace = "i"; break; // LATIN SMALL LETTER I ACUTE
          case '\u00ee': replace = "i"; break; // LATIN SMALL LETTER I CIRCUMFLEX
          case '\u00ef': replace = "i"; break; // LATIN SMALL LETTER I DIAERESIS
          case '\u00f2': replace = "o"; break; // LATIN SMALL LETTER O GRAVE
          case '\u00f3': replace = "o"; break; // LATIN SMALL LETTER O ACUTE
          case '\u00f4': replace = "o"; break; // LATIN SMALL LETTER O CIRCUMFLEX
          case '\u00f6': replace = "o"; break; // LATIN SMALL LETTER O DIAERESIS
          case '\u00f9': replace = "u"; break; // LATIN SMALL LETTER U GRAVE
          case '\u00fa': replace = "u"; break; // LATIN SMALL LETTER U ACUTE
          case '\u00fb': replace = "u"; break; // LATIN SMALL LETTER U CIRCUMFLEX
          case '\u00fc': replace = "u"; break; // LATIN SMALL LETTER U DIAERESIS
          case '\u0100': replace = "A"; break; // LATIN CAPITAL LETTER A MACRON
          case '\u0101': replace = "a"; break; // LATIN SMALL LETTER A MACRON
          case '\u0102': replace = "A"; break; // LATIN CAPITAL LETTER A BREVE
          case '\u0103': replace = "a"; break; // LATIN SMALL LETTER A BREVE
          case '\u0112': replace = "E"; break; // LATIN CAPITAL LETTER E MACRON
          case '\u0113': replace = "e"; break; // LATIN SMALL LETTER E MACRON
          case '\u0114': replace = "E"; break; // LATIN CAPITAL LETTER E BREVE
          case '\u0115': replace = "e"; break; // LATIN SMALL LETTER E BREVE
          case '\u0118': replace = "Ae"; break; // LATIN CAPITAL LETTER E OGONEK
          case '\u0119': replace = "ae"; break; // LATIN SMALL LETTER E OGONEK
          case '\u012a': replace = "I"; break; // LATIN CAPITAL LETTER I MACRON
          case '\u012b': replace = "i"; break; // LATIN SMALL LETTER I MACRON
          case '\u012c': replace = "I"; break; // LATIN CAPITAL LETTER I BREVE
          case '\u012d': replace = "i"; break; // LATIN SMALL LETTER I BREVE
          case '\u014c': replace = "O"; break; // LATIN CAPITAL LETTER O MACRON
          case '\u014d': replace = "o"; break; // LATIN SMALL LETTER O MACRON
          case '\u014e': replace = "O"; break; // LATIN CAPITAL LETTER O BREVE
          case '\u014f': replace = "o"; break; // LATIN SMALL LETTER O BREVE
          case '\u0152': replace = "Oe"; break; // LATIN CAPITAL LETTER O E
          case '\u0153': replace = "oe"; break; // LATIN SMALL LETTER O E
          case '\u016a': replace = "U"; break; // LATIN CAPITAL LETTER U MACRON
          case '\u016b': replace = "u"; break; // LATIN SMALL LETTER U MACRON
          case '\u016c': replace = "U"; break; // LATIN CAPITAL LETTER U BREVE
          case '\u016d': replace = "u"; break; // LATIN SMALL LETTER U BREVE
          case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
          case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
          // new in MPDL project by J. Willenborg
          case '\u1e8d': replace = "e"; break; // LATIN SMALL LETTER E WITH TILDE
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    // new Mpdl code: added by J. Willenborg: most of the latin replacements also in english
    } else if (language.equals("en")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u0300': replace = ""; break; // COMBINING GRAVE ACCENT
          case '\u0301': replace = ""; break; // COMBINING ACCUTE ACCENT
          case '\u0302': replace = ""; break; // COMBINING CIRCUMFLEX ACCENT
    
          case '\u00c0': replace = "A"; break; // LATIN CAPITAL LETTER A GRAVE
          case '\u00c1': replace = "A"; break; // LATIN CAPITAL LETTER A ACUTE
          case '\u00c2': replace = "A"; break; // LATIN CAPITAL LETTER A CIRCUMFLEX
          case '\u00c4': replace = "A"; break; // LATIN CAPITAL LETTER A DIAERESIS
          case '\u00c6': replace = "Ae"; break; // LATIN CAPITAL LETTER A E
          case '\u00c7': replace = "C"; break; // LATIN CAPITAL LETTER C CEDILLA
          case '\u00c8': replace = "E"; break; // LATIN CAPITAL LETTER E GRAVE
          case '\u00c9': replace = "E"; break; // LATIN CAPITAL LETTER E ACUTE
          case '\u00ca': replace = "E"; break; // LATIN CAPITAL LETTER E CIRCUMFLEX
          case '\u00cb': replace = "E"; break; // LATIN CAPITAL LETTER E DIAERESIS
          case '\u00cc': replace = "I"; break; // LATIN CAPITAL LETTER I GRAVE;
          case '\u00cd': replace = "I"; break; // LATIN CAPITAL LETTER I ACUTE
          case '\u00ce': replace = "I"; break; // LATIN CAPITAL LETTER I CIRCUMFLEX
          case '\u00cf': replace = "I"; break; // LATIN CAPITAL LETTER I DIAERESIS
          case '\u00d2': replace = "O"; break; // LATIN CAPITAL LETTER O GRAVE
          case '\u00d3': replace = "O"; break; // LATIN CAPITAL LETTER O ACUTE
          case '\u00d4': replace = "O"; break; // LATIN CAPITAL LETTER O CIRCUMFLEX
          case '\u00d6': replace = "O"; break; // LATIN CAPITAL LETTER O DIAERESIS
          case '\u00d9': replace = "U"; break; // LATIN CAPITAL LETTER U GRAVE
          case '\u00da': replace = "U"; break; // LATIN CAPITAL LETTER U ACUTE
          case '\u00db': replace = "U"; break; // LATIN CAPITAL LETTER U CIRCUMFLEX
          case '\u00dc': replace = "U"; break; // LATIN CAPITAL LETTER U DIAERESIS
          case '\u00e0': replace = "a"; break; // LATIN SMALL LETTER A GRAVE
          case '\u00e1': replace = "a"; break; // LATIN SMALL LETTER A ACUTE
          case '\u00e2': replace = "a"; break; // LATIN SMALL LETTER A CIRCUMFLEX
          case '\u00e4': replace = "a"; break; // LATIN SMALL LETTER A DIAERESIS
          case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
          case '\u00e7': replace = "c"; break; // LATIN SMALL LETTER C CEDILLA
          case '\u00e8': replace = "e"; break; // LATIN SMALL LETTER E GRAVE
          case '\u00e9': replace = "e"; break; // LATIN SMALL LETTER E ACUTE
          case '\u00ea': replace = "e"; break; // LATIN SMALL LETTER E CIRCUMFLEX
          case '\u00eb': replace = "e"; break; // LATIN SMALL LETTER E DIAERESIS
          case '\u00ec': replace = "i"; break; // LATIN SMALL LETTER I GRAVE
          case '\u00ed': replace = "i"; break; // LATIN SMALL LETTER I ACUTE
          case '\u00ee': replace = "i"; break; // LATIN SMALL LETTER I CIRCUMFLEX
          case '\u00ef': replace = "i"; break; // LATIN SMALL LETTER I DIAERESIS
          case '\u00f2': replace = "o"; break; // LATIN SMALL LETTER O GRAVE
          case '\u00f3': replace = "o"; break; // LATIN SMALL LETTER O ACUTE
          case '\u00f4': replace = "o"; break; // LATIN SMALL LETTER O CIRCUMFLEX
          case '\u00f6': replace = "o"; break; // LATIN SMALL LETTER O DIAERESIS
          case '\u00f9': replace = "u"; break; // LATIN SMALL LETTER U GRAVE
          case '\u00fa': replace = "u"; break; // LATIN SMALL LETTER U ACUTE
          case '\u00fb': replace = "u"; break; // LATIN SMALL LETTER U CIRCUMFLEX
          case '\u00fc': replace = "u"; break; // LATIN SMALL LETTER U DIAERESIS
          case '\u0100': replace = "A"; break; // LATIN CAPITAL LETTER A MACRON
          case '\u0101': replace = "a"; break; // LATIN SMALL LETTER A MACRON
          case '\u0102': replace = "A"; break; // LATIN CAPITAL LETTER A BREVE
          case '\u0103': replace = "a"; break; // LATIN SMALL LETTER A BREVE
          case '\u0112': replace = "E"; break; // LATIN CAPITAL LETTER E MACRON
          case '\u0113': replace = "e"; break; // LATIN SMALL LETTER E MACRON
          case '\u0114': replace = "E"; break; // LATIN CAPITAL LETTER E BREVE
          case '\u0115': replace = "e"; break; // LATIN SMALL LETTER E BREVE
          case '\u0118': replace = "Ae"; break; // LATIN CAPITAL LETTER E OGONEK
          case '\u0119': replace = "ae"; break; // LATIN SMALL LETTER E OGONEK
          case '\u012a': replace = "I"; break; // LATIN CAPITAL LETTER I MACRON
          case '\u012b': replace = "i"; break; // LATIN SMALL LETTER I MACRON
          case '\u012c': replace = "I"; break; // LATIN CAPITAL LETTER I BREVE
          case '\u012d': replace = "i"; break; // LATIN SMALL LETTER I BREVE
          case '\u014c': replace = "O"; break; // LATIN CAPITAL LETTER O MACRON
          case '\u014d': replace = "o"; break; // LATIN SMALL LETTER O MACRON
          case '\u014e': replace = "O"; break; // LATIN CAPITAL LETTER O BREVE
          case '\u014f': replace = "o"; break; // LATIN SMALL LETTER O BREVE
          case '\u0152': replace = "Oe"; break; // LATIN CAPITAL LETTER O E
          case '\u0153': replace = "oe"; break; // LATIN SMALL LETTER O E
          case '\u016a': replace = "U"; break; // LATIN CAPITAL LETTER U MACRON
          case '\u016b': replace = "u"; break; // LATIN SMALL LETTER U MACRON
          case '\u016c': replace = "U"; break; // LATIN CAPITAL LETTER U BREVE
          case '\u016d': replace = "u"; break; // LATIN SMALL LETTER U BREVE
          case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
          case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
          // new in MPDL project by J. Willenborg
          case '\u1e8d': replace = "e"; break; // LATIN SMALL LETTER E WITH TILDE
          // by Malcolm
          case '\u00ad': break; // soft hyphen
          case '\u2329': break; // BRA
          case '\u232a': break; // KET
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("fr")) {
      // new Mpdl code: added by J. Willenborg: some of Malcolms code did not work without errors so it has to be taken away
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
          case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
          case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
          case '\u00ad': break; // soft hyphen
          case '-': break;
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("de")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u00c4': replace = "Ae"; break;
          case '\u00d6': replace = "Oe"; break;
          case '\u00dc': replace = "Ue"; break;
          case '\u00df': replace = "ss"; break;
          case '\u00e4': replace = "ae"; break;
          case '\u00f6': replace = "oe"; break;
          case '\u00fc': replace = "ue"; break;
          case '\u00e9': replace = "e"; break;
          case '\u00ad': break; // soft hyphen
          case '-': break;
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("zh")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u00b9': replace = "1"; break;
          case '\u00b2': replace = "2"; break;
          case '\u00b3': replace = "3"; break;
          case '\u2074': replace = "4"; break;
          case '\u2075': replace = "5"; break;
          // original by Malcolm Hyman: with the following replacements // TODO uncomment these 3 lines
          // case '\u3000': replace = " "; break;
          // case '\u3001': replace = ","; break;
          // case '\u3002': replace = "."; break;
          // case '\u200b': break; // BREAKS EVERYTHING!
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("akk") ||
       language.equals("qam") ||
       language.equals("qpc") ||
       language.equals("elx") ||
       language.equals("sux") ||
       language.equals("hit") ||
       language.equals("qhu") ||
       language.equals("peo") ||
       language.equals("uga") ||
       language.equals("ura") ||
       language.equals("qcu")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      char last = '\u0000';
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        c = Character.toLowerCase(c);
        String replace = new String();
        switch (c) {
          case '{': replace += "-"; break;
          case '}': replace += "-"; break;
          // These are from PSD::ATF::Unicode by Steve Tinney
          case '\u0161': replace += "sz"; break;
          case '\u1e63': replace += "s,"; break;
          case '\u1e6d': replace += "t,"; break;
          case '\u014b': replace += "j"; break;
          case '\u015b': replace += "s'"; break;
          case '\u2080': replace += "0"; break;
          case '\u2081': replace += "1"; break;
          case '\u2082': replace += "2"; break;
          case '\u2083': replace += "3"; break;
          case '\u2084': replace += "4"; break;
          case '\u2085': replace += "5"; break;
          case '\u2086': replace += "6"; break;
          case '\u2087': replace += "7"; break;
          case '\u2088': replace += "8"; break;
          case '\u2089': replace += "9"; break;
    
          case 'c': // shin (except where used as modifier)
            if ((i > 0) && ((last == '~') || (last == '@')))
              replace += "c";
            else replace += "sz";
              break;
          default: replace += c; break;
        }
        // suppress grapheme boundary before or after word boundary
        if (replace.equals("-")) {
          if ((i + 1 == s.length()) || (s.charAt(i + 1) == ' ') || (i == 0) || (s.charAt(i - 1) == ' ')) 
            replace = "";
        }
        last = c;
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("el") || language.equals("grc")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u03c2': replace = "\u03c3"; break; // GREEK SMALL LETTER FINAL SIGMA
          case '<': break;
          case '>': break;
          case '[': break;
          case ']': break;
          case '1': break;
          case '2': break;
          case '\u03ac': replace = "\u1f71"; break;
          case '\u03ad': replace = "\u1f73"; break;
          case '\u03ae': replace = "\u1f75"; break;
          case '\u03af': replace = "\u1f77"; break;
          case '\u03cc': replace = "\u1f79"; break;
          case '\u03cd': replace = "\u1f7b"; break;
          case '\u03ce': replace = "\u1f7d"; break;
          case '-': break; // same treatment as soft hyphen
          case '\u00ad': break; // soft hyphen
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else if (language.equals("el_atonic")) {
      StringBuffer buf = new StringBuffer();
      int n = 0;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        String replace = new String();
        switch (c) {
          case '\u03c2': replace = "\u03c3"; break; // GREEK SMALL LETTER FINAL SIGMA
          // map characters with diacritics to their plain equivalent
          // cf. <code>BetaCode.java</code>
          case '\u03aa': replace = "\u0399"; break;
          case '\u03ab': replace = "\u03a5"; break;
          case '\u03ac': replace = "\u0381"; break;
          case '\u03ad': replace = "\u0385"; break;
          case '\u03ae': replace = "\u0387"; break;
          case '\u03af': replace = "\u0389"; break;
          case '\u03ca': replace = "\u03b9"; break;
          case '\u03cb': replace = "\u03c5"; break;
          case '\u03cc': replace = "\u03bf"; break;
          case '\u03cd': replace = "\u03c5"; break;
          case '\u03ce': replace = "\u03c9"; break;
          case '\u1f00': replace = "\u03b1"; break;
          case '\u1f01': replace = "\u03b1"; break;
          case '\u1f02': replace = "\u03b1"; break;
          case '\u1f03': replace = "\u03b1"; break;
          case '\u1f04': replace = "\u03b1"; break;
          case '\u1f05': replace = "\u03b1"; break;
          case '\u1f06': replace = "\u03b1"; break;
          case '\u1f07': replace = "\u03b1"; break;
          case '\u1f08': replace = "\u0391"; break;
          case '\u1f09': replace = "\u0391"; break;
          case '\u1f0a': replace = "\u0391"; break;
          case '\u1f0b': replace = "\u0391"; break;
          case '\u1f0c': replace = "\u0391"; break;
          case '\u1f0d': replace = "\u0391"; break;
          case '\u1f0e': replace = "\u0391"; break;
          case '\u1f0f': replace = "\u0391"; break;
          case '\u1f10': replace = "\u03b5"; break;
          case '\u1f11': replace = "\u03b5"; break;
          case '\u1f12': replace = "\u03b5"; break;
          case '\u1f13': replace = "\u03b5"; break;
          case '\u1f14': replace = "\u03b5"; break;
          case '\u1f15': replace = "\u03b5"; break;
          case '\u1f18': replace = "\u0395"; break;
          case '\u1f19': replace = "\u0395"; break;
          case '\u1f1a': replace = "\u0395"; break;
          case '\u1f1b': replace = "\u0395"; break;
          case '\u1f1c': replace = "\u0395"; break;
          case '\u1f1d': replace = "\u0395"; break;
          case '\u1f20': replace = "\u03b7"; break;
          case '\u1f21': replace = "\u03b7"; break;
          case '\u1f22': replace = "\u03b7"; break;
          case '\u1f23': replace = "\u03b7"; break;
          case '\u1f24': replace = "\u03b7"; break;
          case '\u1f25': replace = "\u03b7"; break;
          case '\u1f26': replace = "\u03b7"; break;
          case '\u1f27': replace = "\u03b7"; break;
          case '\u1f28': replace = "\u0397"; break;
          case '\u1f29': replace = "\u0397"; break;
          case '\u1f2a': replace = "\u0397"; break;
          case '\u1f2b': replace = "\u0397"; break;
          case '\u1f2c': replace = "\u0397"; break;
          case '\u1f2d': replace = "\u0397"; break;
          case '\u1f2e': replace = "\u0397"; break;
          case '\u1f2f': replace = "\u0397"; break;
          case '\u1f30': replace = "\u03b9"; break;
          case '\u1f31': replace = "\u03b9"; break;
          case '\u1f32': replace = "\u03b9"; break;
          case '\u1f33': replace = "\u03b9"; break;
          case '\u1f34': replace = "\u03b9"; break;
          case '\u1f35': replace = "\u03b9"; break;
          case '\u1f36': replace = "\u03b9"; break;
          case '\u1f37': replace = "\u03b9"; break;
          case '\u1f38': replace = "\u0399"; break;
          case '\u1f39': replace = "\u0399"; break;
          case '\u1f3a': replace = "\u0399"; break;
          case '\u1f3b': replace = "\u0399"; break;
          case '\u1f3c': replace = "\u0399"; break;
          case '\u1f3d': replace = "\u0399"; break;
          case '\u1f3e': replace = "\u0399"; break;
          case '\u1f3f': replace = "\u0399"; break;
          case '\u1f40': replace = "\u03bf"; break;
          case '\u1f41': replace = "\u03bf"; break;
          case '\u1f42': replace = "\u03bf"; break;
          case '\u1f43': replace = "\u03bf"; break;
          case '\u1f44': replace = "\u03bf"; break;
          case '\u1f45': replace = "\u03bf"; break;
          case '\u1f48': replace = "\u039f"; break;
          case '\u1f49': replace = "\u039f"; break;
          case '\u1f4a': replace = "\u039f"; break;
          case '\u1f4b': replace = "\u039f"; break;
          case '\u1f4c': replace = "\u039f"; break;
          case '\u1f4d': replace = "\u039f"; break;
          case '\u1f50': replace = "\u03c5"; break;
          case '\u1f51': replace = "\u03c5"; break;
          case '\u1f52': replace = "\u03c5"; break;
          case '\u1f53': replace = "\u03c5"; break;
          case '\u1f54': replace = "\u03c5"; break;
          case '\u1f55': replace = "\u03c5"; break;
          case '\u1f56': replace = "\u03c5"; break;
          case '\u1f57': replace = "\u03c5"; break;
          case '\u1f58': replace = "\u03a5"; break;
          case '\u1f59': replace = "\u03a5"; break;
          case '\u1f5a': replace = "\u03a5"; break;
          case '\u1f5b': replace = "\u03a5"; break;
          case '\u1f5c': replace = "\u03a5"; break;
          case '\u1f5d': replace = "\u03a5"; break;
          case '\u1f5e': replace = "\u03a5"; break;
          case '\u1f5f': replace = "\u03a5"; break;
          case '\u1f60': replace = "\u03c9"; break;
          case '\u1f61': replace = "\u03c9"; break;
          case '\u1f62': replace = "\u03c9"; break;
          case '\u1f63': replace = "\u03c9"; break;
          case '\u1f64': replace = "\u03c9"; break;
          case '\u1f65': replace = "\u03c9"; break;
          case '\u1f66': replace = "\u03c9"; break;
          case '\u1f67': replace = "\u03c9"; break;
          case '\u1f68': replace = "\u03a9"; break;
          case '\u1f69': replace = "\u03a9"; break;
          case '\u1f6a': replace = "\u03a9"; break;
          case '\u1f6b': replace = "\u03a9"; break;
          case '\u1f6c': replace = "\u03a9"; break;
          case '\u1f6d': replace = "\u03a9"; break;
          case '\u1f6e': replace = "\u03a9"; break;
          case '\u1f6f': replace = "\u03a9"; break;
          case '\u1f70': replace = "\u03b1"; break;
          case '\u1f71': replace = "\u03b1"; break;
          case '\u1f72': replace = "\u03b5"; break;
          case '\u1f73': replace = "\u03b5"; break;
          case '\u1f74': replace = "\u03b7"; break;
          case '\u1f75': replace = "\u03b7"; break;
          case '\u1f76': replace = "\u03b9"; break;
          case '\u1f77': replace = "\u03b9"; break;
          case '\u1f78': replace = "\u03bf"; break;
          case '\u1f79': replace = "\u03bf"; break;
          case '\u1f7a': replace = "\u03c5"; break;
          case '\u1f7b': replace = "\u03c5"; break;
          case '\u1f7c': replace = "\u03c9"; break;
          case '\u1f7d': replace = "\u03c9"; break;
          case '\u1f80': replace = "\u03b1"; break;
          case '\u1f81': replace = "\u03b1"; break;
          case '\u1f82': replace = "\u03b1"; break;
          case '\u1f83': replace = "\u03b1"; break;
          case '\u1f84': replace = "\u03b1"; break;
          case '\u1f85': replace = "\u03b1"; break;
          case '\u1f86': replace = "\u03b1"; break;
          case '\u1f87': replace = "\u03b1"; break;
          case '\u1f88': replace = "\u0391"; break;
          case '\u1f89': replace = "\u0391"; break;
          case '\u1f8a': replace = "\u0391"; break;
          case '\u1f8b': replace = "\u0391"; break;
          case '\u1f8c': replace = "\u0391"; break;
          case '\u1f8d': replace = "\u0391"; break;
          case '\u1f8e': replace = "\u0391"; break;
          case '\u1f8f': replace = "\u0391"; break;
          case '\u1f90': replace = "\u03b7"; break;
          case '\u1f91': replace = "\u03b7"; break;
          case '\u1f92': replace = "\u03b7"; break;
          case '\u1f93': replace = "\u03b7"; break;
          case '\u1f94': replace = "\u03b7"; break;
          case '\u1f95': replace = "\u03b7"; break;
          case '\u1f96': replace = "\u03b7"; break;
          case '\u1f97': replace = "\u03b7"; break;
          case '\u1f98': replace = "\u0397"; break;
          case '\u1f99': replace = "\u0397"; break;
          case '\u1f9a': replace = "\u0397"; break;
          case '\u1f9b': replace = "\u0397"; break;
          case '\u1f9c': replace = "\u0397"; break;
          case '\u1f9d': replace = "\u0397"; break;
          case '\u1f9e': replace = "\u0397"; break;
          case '\u1f9f': replace = "\u0397"; break;
          case '\u1fa0': replace = "\u03c9"; break;
          case '\u1fa1': replace = "\u03c9"; break;
          case '\u1fa2': replace = "\u03c9"; break;
          case '\u1fa3': replace = "\u03c9"; break;
          case '\u1fa4': replace = "\u03c9"; break;
          case '\u1fa5': replace = "\u03c9"; break;
          case '\u1fa6': replace = "\u03c9"; break;
          case '\u1fa7': replace = "\u03c9"; break;
          case '\u1fa8': replace = "\u03a9"; break;
          case '\u1fa9': replace = "\u03a9"; break;
          case '\u1faa': replace = "\u03a9"; break;
          case '\u1fab': replace = "\u03a9"; break;
          case '\u1fac': replace = "\u03a9"; break;
          case '\u1fad': replace = "\u03a9"; break;
          case '\u1fae': replace = "\u03a9"; break;
          case '\u1faf': replace = "\u03a9"; break;
          case '\u1fb2': replace = "\u03b1"; break;
          case '\u1fb3': replace = "\u03b1"; break;
          case '\u1fb4': replace = "\u03b1"; break;
          case '\u1fb6': replace = "\u03b1"; break;
          case '\u1fb7': replace = "\u03b1"; break;
          case '\u1fba': replace = "\u0391"; break;
          case '\u1fbb': replace = "\u0391"; break;
          case '\u1fbc': replace = "\u0391"; break;
          case '\u1fc2': replace = "\u03b7"; break;
          case '\u1fc3': replace = "\u03b7"; break;
          case '\u1fc4': replace = "\u03b7"; break;
          case '\u1fc6': replace = "\u03b7"; break;
          case '\u1fc7': replace = "\u03b7"; break;
          case '\u1fca': replace = "\u0397"; break;
          case '\u1fcb': replace = "\u0397"; break;
          case '\u1fcc': replace = "\u0397"; break;
          case '\u1fd2': replace = "\u03b9"; break;
          case '\u1fd3': replace = "\u03b9"; break;
          case '\u1fd6': replace = "\u03b9"; break;
          case '\u1fd7': replace = "\u03b9"; break;
          case '\u1fda': replace = "\u0399"; break;
          case '\u1fdb': replace = "\u039f"; break;
          case '\u1fe2': replace = "\u03c5"; break;
          case '\u1fe3': replace = "\u03c5"; break;
          case '\u1fe4': replace = "\u03c1"; break;
          case '\u1fe5': replace = "\u03c1"; break;
          case '\u1fe6': replace = "\u03c5"; break;
          case '\u1fe7': replace = "\u03c5"; break;
          case '\u1fea': replace = "\u03a5"; break;
          case '\u1feb': replace = "\u03a5"; break;
          case '\u1fec': replace = "\u03a1"; break;
          case '\u1ff2': replace = "\u03c9"; break;
          case '\u1ff3': replace = "\u03c9"; break;
          case '\u1ff4': replace = "\u03c9"; break;
          case '\u1ff6': replace = "\u03c9"; break;
          case '\u1ff7': replace = "\u03c9"; break;
          case '\u1ff8': replace = "\u039f"; break;
          case '\u1ff9': replace = "\u039f"; break;
          case '\u1ffa': replace = "\u03a9"; break;
          case '\u1ffb': replace = "\u03a9"; break;
          case '\u1ffc': replace = "\u03a9"; break;
    
          case '<': break;
          case '>': break;
          case '-': break; // same treatment as soft hyphen
          case '\u00ad': break; // soft hyphen
          default: replace += c; break;
        }
        buf.append(replace);
        // update offsets if replacement is a different length
        if (offsets != null) {
          int r = replace.length();
          if (r == 0) 
            this.offsets = arrayKill(this.offsets, i - n);
          else if (r == 2)
            this.offsets = arrayInsert(this.offsets, i - n + 1, this.offsets[i - n], r - 1);
          n += 1 - r;
        }
      }
      return buf.toString();
    } else {      // unknown or no language
      return s;
    }
  }
  
  public String deNormalizeToRegExpr(String s) {
    // TODO all characters in all languages
    if (language.equals("la") || language.equals("lat")) {
      StringBuffer buf = new StringBuffer();
      if (s.indexOf("ae") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ae", "\u0119");
        String str2 = s;
        str2 = str2.replaceAll("ae", "\u00c6");
        String str3 = s;
        str3 = str3.replaceAll("ae", "\u00e6");
        buf.append(str1 + "|" + str2 + "|" + str3 + "|");
      }
      if (s.indexOf("oe") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("oe", "\u0152");
        String str2 = s;
        str2 = str2.replaceAll("oe", "\u0153");
        buf.append(str1 + "|" + str2 + "|");
      }
      if (s.indexOf("ss") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ss", "\u00df");
        buf.append(str1 + "|");
      }
      boolean beginWord = true;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        if (! beginWord) 
          c = Character.toLowerCase(c);
        beginWord = Character.isWhitespace(c);
        String replace = new String();
        switch (c) {
          case 'a': replace = "[a\u00c0\u00c1\u00c2\u00c4\u00e0\u00e1\u00e2\u00e4]"; break; 
          case 'c': replace = "[c\u00c7\u00e7]"; break;
          case 'e': replace = "[e\u00c8\u00c9\u00ca\u00cb\u00e8\u00e9\u00ea\u00eb\u0113\u0115\u1ebd]"; break; 
          case 'i': replace = "[ij\u00cc\u00cd\u00ce\u00cf\u00ec\u00ed\u00ee\u00ef\u012a\u012b\u012c\u012d]"; break;
          case 'o': replace = "[o\u00d2\u00d3\u00d4\u00d6\u00f2\u00f3\u00f4\u00f6\u014c\u014d\u014e\u014f]"; break; 
          case 'u': replace = "[uv\u00d9\u00da\u00db\u00dc\u00f9\u00fa\u00fb\u00fc\u016a\u016b\u016c\u016d]"; break; 
          case 's': replace = "[s\u017f]"; break; 
          default: replace += c; break;
        }
        buf.append(replace);
      }
      return buf.toString();
    } else if (language.equals("en")) {
      StringBuffer buf = new StringBuffer();
      if (s.indexOf("ae") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ae", "\u0119");
        String str2 = s;
        str2 = str2.replaceAll("ae", "\u00c6");
        String str3 = s;
        str3 = str3.replaceAll("ae", "\u00e6");
        buf.append(str1 + "|" + str2 + "|" + str3 + "|");
      }
      if (s.indexOf("oe") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("oe", "\u0152");
        String str2 = s;
        str2 = str2.replaceAll("oe", "\u0153");
        buf.append(str1 + "|" + str2 + "|");
      }
      if (s.indexOf("ss") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ss", "\u00df");
        buf.append(str1 + "|");
      }
      boolean beginWord = true;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        if (! beginWord) 
          c = Character.toLowerCase(c);
        beginWord = Character.isWhitespace(c);
        String replace = new String();
        switch (c) {
          case 'a': replace = "[a\u00c0\u00c1\u00c2\u00c4\u00e0\u00e1\u00e2\u00e4]"; break; 
          case 'c': replace = "[c\u00c7\u00e7]"; break;
          case 'e': replace = "[e\u00c8\u00c9\u00ca\u00cb\u00e8\u00e9\u00ea\u00eb\u0113\u0115\u1e8d]"; break; 
          case 'i': replace = "[i\u00cc\u00cd\u00ce\u00cf\u00ec\u00ed\u00ee\u00ef\u012a\u012b\u012c\u012d]"; break;
          case 'o': replace = "[o\u00d2\u00d3\u00d4\u00d6\u00f2\u00f3\u00f4\u00f6\u014c\u014d\u014e\u014f]‚"; break; 
          case 'u': replace = "[u\u00d9\u00da\u00db\u00dc\u00f9\u00fa\u00fb\u00fc\u016a\u016b\u016c\u016d]"; break; 
          case 's': replace = "[s\u017f]"; break; 
          default: replace += c; break;
        }
        buf.append(replace);
      }
      return buf.toString();
    } else if (language.equals("de")) {
      StringBuffer buf = new StringBuffer();
      if (s.indexOf("ss") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ss", "\u00df");
        buf.append(str1 + "|");
      }
      if (s.indexOf("ae") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ae", "\u00e4");
        buf.append(str1 + "|");
      }
      if (s.indexOf("oe") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("oe", "\u00f6");
        buf.append(str1 + "|");
      }
      if (s.indexOf("ue") != -1) {
        String str1 = s;
        str1 = str1.replaceAll("ue", "\u00fc");
        buf.append(str1 + "|");
      }
      boolean beginWord = true;
      for (int i = 0; i < s.length(); i++) {
        char c = s.charAt(i);
        if (! beginWord) 
          c = Character.toLowerCase(c);
        beginWord = Character.isWhitespace(c);
        String replace = new String();
        switch (c) {
          case 'e': replace = "[e\u00e9]"; break; 
          default: replace += c; break;
        }
        buf.append(replace);
      }
      return buf.toString();
    } else {      // unknown or no language
      return s;
    }
  }
  
  /**
   * Returns the offset table.
   *
   * @return             offset table
   */
  public int[] getOffsetTable() {
    return offsets;
  }

  /**
   * Returns a copy of an integer array with the element at
   * <code>index</code> removed ("killed").
   *
   * @param array        integer array
   * @param index        index of element to remove
   */
  static private int[] arrayKill(int[] array, int index) {
    int[] newArray = new int[array.length - 1];
    System.arraycopy(array, 0, newArray, 0, index);
    System.arraycopy(array, index + 1, newArray, index, array.length - index - 1);
    return newArray;
  }

  /**
   * Returns a copy of an integer array with <code>count</code> elements
   * inserted at <code>index</code>.
   *
   * @param array        integer array
   * @param index        index to insert new elements
   * @param value        value to insert into new slots
   * @param count        number of new slots to insert
   */
  static private int[] arrayInsert(int[] array, int index, int value, int count) {
    int[] newArray = new int[array.length + count];
    System.arraycopy(array, 0, newArray, 0, index);
    for (int i = 0; i < count; i++) newArray[index + i] = value;
    System.arraycopy(array, index, newArray, index + count, array.length - index);
    return newArray;
  }

  /**
   * We provide <code>main()</code> so that our services will be available
   * outside Java (i.e., so we can run as a Un*x-style filter).
   */
  static public void main(String[] argv) throws ApplicationException {
    if (argv.length != 1) {
      System.err.println("You must specify a language.");
      System.exit(1);
    }
    String rec;
    StringBuffer buf = new StringBuffer();
    BufferedReader bin = null;
    try {
      bin = new BufferedReader(new InputStreamReader(System.in, "UTF-8"));
      while ((rec = bin.readLine()) != null) 
        buf.append(rec + "\n");
    }
    catch (UnsupportedEncodingException e) {
      System.err.println(e);
      System.exit(1);
    } catch (IOException e) {
      System.err.println(e);
      System.exit(1);
    }
    MpdlNormalizer orth = new MpdlNormalizer(argv[0]);
    System.out.print(orth.normalize(buf.toString()));
  }
}