import java.io.*;

/**
 * This class provides pluggable rules for orthographic normalization, with
 * offset tracking.
 *
 * @author  Malcolm D. Hyman
 * @version $2007-06-12 18:58:58 mdh$
 */
public class OrthographicRules {
    private String ruleset;
    public int[] offsets;

    static final public String IT_VOWELS    = "AEIOUaeiou" +
	                                      "\u00c6\u00e6" + // AE ligatures
	                                      "\u0152\u0153";  // OE ligatures
    static final public String IT_CONS      = "BCDFGHKLMNPQRSTVWXZ" +
	                                      "bcdfghklmnpqrstvwxz" +
	                                      "\u017f\u00df";  // long/sharp S


    /**
     * Constructor.
     *
     * @param ruleset      name of rule set to apply
     */
    public OrthographicRules(String ruleset) {
	this.ruleset = ruleset;
    }
	
    /**
     * Applies the normalization rules in <code>ruleset</code> to
     * <code>s</code>, with offset tracking.<p>
     *
     * <strong>WARNING:</strong>
     * Arboreal will not work properly if a normalization substitution
     * replaces a source character with more than two target characters!
     * This is simply a BUG, and should be fixed. Fortunately, however,
     * one does not often need such a replacement.<p>
     *
     * FIXME: If the orthographic rules eliminate all characters in a 
     * word, word counting in ContentRenderPane will not work correctly!
     *
     * @param s            source string
     * @param offsets      character offset table
     * @return             normalized string
     */
    public String normalize(String s, int[] offsets) {
	this.offsets = offsets;
	if (ruleset.equals("la") || ruleset.equals("lat")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		if (! beginWord) c = Character.toLowerCase(c);
		beginWord = Character.isWhitespace(c);
		String replace = new String();

		// names in comments are from Unicode Standard
		switch (c) {
		case 'j': replace = "i"; break;
		case 'v': replace = "u"; break;

		/*
		 * Linguistic note: /u/ and /v/ are rarely phonemic
		 * in Latin, as in alui 's/he nourished' vs.
		 * alvi 'of a belly', volui 's/he wished' or 'it rolled'
		 * vs. volvi 'to be rolled', (in)seruit 's/he joined
		 * together' vs. (in)servit 's/he serves'.
		 */

		case 'q':
		    if ((i < s.length() - 1) && (s.charAt(i + 1) == ';'))
			replace = "qu";
		    else
			replace = "q";
		    break;
 		case ';':
 		    if ((i > 0) && (s.charAt(i - 1) == 'q'))
 			replace = "e";
		    else if ((i == 0) || ! Character.isLetter(s.charAt(i - 1)))
			replace = ";";
 		    else
 			replace = "";
 		    break;

		case '\u0300': replace = ""; break; // COMBINING GRAVE ACCENT
		case '\u0301': replace = ""; break; // COMBINING ACCUTE ACCENT
		case '\u0302': replace = ""; break; // COMBINING CIRCUMFLEX ACCENT

		case '\u00c0': replace = "A"; break; // LATIN CAPITAL LETTER A GRAVE
		case '\u00c1': replace = "A"; break; // LATIN CAPITAL LETTER A ACUTE
		case '\u00c2': replace = "A"; break; // LATIN CAPITAL LETTER A CIRCUMFLEX
		case '\u00c4': replace = "A"; break; // LATIN CAPITAL LETTER A DIAERESIS
		case '\u00c6': replace = "Ae"; break; // LATIN CAPITAL LETTER A E
		case '\u00c7': replace = "C"; break; // LATIN CAPITAL LETTER C CEDILLA
		case '\u00c8': replace = "E"; break; // LATIN CAPITAL LETTER E GRAVE
		case '\u00c9': replace = "E"; break; // LATIN CAPITAL LETTER E ACUTE
		case '\u00ca': replace = "E"; break; // LATIN CAPITAL LETTER E CIRCUMFLEX
		case '\u00cb': replace = "E"; break; // LATIN CAPITAL LETTER E DIAERESIS
		case '\u00cc': replace = "I"; break; // LATIN CAPITAL LETTER I GRAVE;
		case '\u00cd': replace = "I"; break; // LATIN CAPITAL LETTER I ACUTE
		case '\u00ce': replace = "I"; break; // LATIN CAPITAL LETTER I CIRCUMFLEX
		case '\u00cf': replace = "I"; break; // LATIN CAPITAL LETTER I DIAERESIS
		case '\u00d2': replace = "O"; break; // LATIN CAPITAL LETTER O GRAVE
		case '\u00d3': replace = "O"; break; // LATIN CAPITAL LETTER O ACUTE
		case '\u00d4': replace = "O"; break; // LATIN CAPITAL LETTER O CIRCUMFLEX
		case '\u00d6': replace = "O"; break; // LATIN CAPITAL LETTER O DIAERESIS
		case '\u00d9': replace = "U"; break; // LATIN CAPITAL LETTER U GRAVE
		case '\u00da': replace = "U"; break; // LATIN CAPITAL LETTER U ACUTE
		case '\u00db': replace = "U"; break; // LATIN CAPITAL LETTER U CIRCUMFLEX
		case '\u00dc': replace = "U"; break; // LATIN CAPITAL LETTER U DIAERESIS
		case '\u00e0': replace = "a"; break; // LATIN SMALL LETTER A GRAVE
		case '\u00e1': replace = "a"; break; // LATIN SMALL LETTER A ACUTE
		case '\u00e2': replace = "a"; break; // LATIN SMALL LETTER A CIRCUMFLEX
		case '\u00e4': replace = "a"; break; // LATIN SMALL LETTER A DIAERESIS
		case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
		case '\u00e7': replace = "c"; break; // LATIN SMALL LETTER C CEDILLA
		case '\u00e8': replace = "e"; break; // LATIN SMALL LETTER E GRAVE
		case '\u00e9': replace = "e"; break; // LATIN SMALL LETTER E ACUTE
		case '\u00ea': replace = "e"; break; // LATIN SMALL LETTER E CIRCUMFLEX
		case '\u00eb': replace = "e"; break; // LATIN SMALL LETTER E DIAERESIS
		case '\u00ec': replace = "i"; break; // LATIN SMALL LETTER I GRAVE
		case '\u00ed': replace = "i"; break; // LATIN SMALL LETTER I ACUTE
		case '\u00ee': replace = "i"; break; // LATIN SMALL LETTER I CIRCUMFLEX
		case '\u00ef': replace = "i"; break; // LATIN SMALL LETTER I DIAERESIS
		case '\u00f2': replace = "o"; break; // LATIN SMALL LETTER O GRAVE
		case '\u00f3': replace = "o"; break; // LATIN SMALL LETTER O ACUTE
		case '\u00f4': replace = "o"; break; // LATIN SMALL LETTER O CIRCUMFLEX
		case '\u00f6': replace = "o"; break; // LATIN SMALL LETTER O DIAERESIS
		case '\u00f9': replace = "u"; break; // LATIN SMALL LETTER U GRAVE
		case '\u00fa': replace = "u"; break; // LATIN SMALL LETTER U ACUTE
		case '\u00fb': replace = "u"; break; // LATIN SMALL LETTER U CIRCUMFLEX
		case '\u00fc': replace = "u"; break; // LATIN SMALL LETTER U DIAERESIS
		case '\u0100': replace = "A"; break; // LATIN CAPITAL LETTER A MACRON
		case '\u0101': replace = "a"; break; // LATIN SMALL LETTER A MACRON
		case '\u0102': replace = "A"; break; // LATIN CAPITAL LETTER A BREVE
		case '\u0103': replace = "a"; break; // LATIN SMALL LETTER A BREVE
		case '\u0112': replace = "E"; break; // LATIN CAPITAL LETTER E MACRON
		case '\u0113': replace = "e"; break; // LATIN SMALL LETTER E MACRON
		case '\u0114': replace = "E"; break; // LATIN CAPITAL LETTER E BREVE
		case '\u0115': replace = "e"; break; // LATIN SMALL LETTER E BREVE
		case '\u0118': replace = "Ae"; break; // LATIN CAPITAL LETTER E OGONEK
		case '\u0119': replace = "ae"; break; // LATIN SMALL LETTER E OGONEK
		case '\u012a': replace = "I"; break; // LATIN CAPITAL LETTER I MACRON
		case '\u012b': replace = "i"; break; // LATIN SMALL LETTER I MACRON
		case '\u012c': replace = "I"; break; // LATIN CAPITAL LETTER I BREVE
		case '\u012d': replace = "i"; break; // LATIN SMALL LETTER I BREVE
		case '\u014c': replace = "O"; break; // LATIN CAPITAL LETTER O MACRON
		case '\u014d': replace = "o"; break; // LATIN SMALL LETTER O MACRON
		case '\u014e': replace = "O"; break; // LATIN CAPITAL LETTER O BREVE
		case '\u014f': replace = "o"; break; // LATIN SMALL LETTER O BREVE
		case '\u0152': replace = "Oe"; break; // LATIN CAPITAL LETTER O E
		case '\u0153': replace = "oe"; break; // LATIN SMALL LETTER O E
		case '\u016a': replace = "U"; break; // LATIN CAPITAL LETTER U MACRON
		case '\u016b': replace = "u"; break; // LATIN SMALL LETTER U MACRON
		case '\u016c': replace = "U"; break; // LATIN CAPITAL LETTER U BREVE
		case '\u016d': replace = "u"; break; // LATIN SMALL LETTER U BREVE
                case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
                case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
		case '\u00ad': break; // soft hyphen
		case '\u2329': break; // BRA
		case '\u232a': break; // KET
		default: replace += c; break;
		}

		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}
	
	else if (ruleset.equals("it")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		if (! beginWord) c = Character.toLowerCase(c);
		String replace = new String();
	    
		if ((c == 'u') && (i > 0) && (i < s.length() - 1)) {
		    char c_n = s.charAt(i + 1);
		    char c_p;
		    int y = i - 1;
		    while ((((c_p = s.charAt(y)) == '-') ||
			    (c_p == '\u00ad')) && (y-- > 0));
		    if ((IT_VOWELS.indexOf(c_p) > -1) &&
			(IT_VOWELS.indexOf(c_n) > -1)) replace = "v";
		    else replace = "u";
		}
		else if ((c == 'j') && (i > 0)) {
		    char c_p = Character.toLowerCase(s.charAt(i - 1));
		    if (c_p == 'i') replace = "i";
		    else replace = "j";
		}
		else if ((c == 'v') && (i < s.length() - 1)) {
		    char c_p;
		    char c_n;
		    int y = i + 1;
		    while ((((c_n = s.charAt(y)) == '-') ||
			    (c_n == '\u00ad')) && (y++ > 0));
		    if ((beginWord ||
			 (! Character.isLetter(c_p = s.charAt(i - 1))) ||
			 IT_CONS.indexOf(c_p) > -1) &&
			IT_CONS.indexOf(c_n) > -1)
			replace = "u"; else replace = "v";
		}
		else if ((c == 't') && (i < s.length() - 2)) {
		    char c_n = Character.toLowerCase(s.charAt(i + 1));
		    char c_n2 = Character.toLowerCase(s.charAt(i + 2));
		    if ((c_n == 'i') && (c_n2 == 'o'))
			replace = "z";
		    else
			replace  = "t";
                }
		/*
		 * 'h' only appears at the beginning of some pres. indic.
		 * forms of 'avere' -- with the exception of some unimportant
		 * loanwords in modern Italian (which I choose to ignore)
		 */
		else if (c == 'h') {
		    int l = s.length();
		    if ((i < l - 1) &&
			(beginWord ||
			 ((i > 0) &&
			  (! Character.isLetter(s.charAt(i - 1)))))) {
			char c_n = Character.toLowerCase(s.charAt(i + 1));
			if (c_n == 'a') {
			    if (i < l - 2) {
				char c_n2 =
				    Character.toLowerCase(s.charAt(i + 2));
				if (((! Character.isLetter(c_n2)) && //  'ha'
				     (c_n2 != '-') && (c_n2 != '\u00ad')) ||
				    (c_n2 == 'i') || // 'hai'
				    (c_n2 == 'n')) { // 'hanno'
				    replace = "h";
				}
				else {
				    replace = "";
				}
			    }
			    else {
				replace = "h"; // 'ha'
			    }
			}
			else if (c_n == 'o') {
			    if (i < l - 2) {
				char c_n2 =
				    Character.toLowerCase(s.charAt(i + 2));
				if (! Character.isLetter(c_n2)) {
				    replace = "h";
				}
				else {
				    replace = "";
				}
			    }
			    else {
				replace = "h"; // 'ho'
			    }
			}
			else {
			    replace = "";
			}
		    }
		    else {
			replace = "h"; // not word initial (or else isolated)
		    }
		}
		else if ((c == '\u00ad') || (c == '-')) { // (soft) hyphen
		    if ((i < s.length() - 1) &&
			(! Character.isWhitespace(s.charAt(i + 1)))) {
			replace = "";
		    }
		    else {
			replace = "-";
		    }
		}
		else {
                    switch (c) {
                    case '\u00c1': replace = "\u00c0"; break; // LATIN CAPITAL LETTER A ACUTE
                    case '\u00c6': replace = "Ae"; break; // LATIN CAPITAL LETTER A E
                    case '\u00c9': replace = "\u00c8"; break; // LATIN CAPITAL LETTER E ACUTE
                    case '\u00cd': replace = "\u00cc"; break; // LATIN CAPITAL LETTER I ACUTE
                    case '\u00d3': replace = "\u00d2"; break; // LATIN CAPITAL LETTER O ACUTE
                    case '\u00da': replace = "\u00d9"; break; // LATIN CAPITAL LETTER U ACUTE
                    case '\u00e1': replace = "\u00e0"; break; // LATIN SMALL LETTER A ACUTE
                    case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
                    case '\u00e9': replace = "\u00e8"; break; // LATIN SMALL LETTER E ACUTE
                    case '\u00ed': replace = "\u00ec"; break; // LATIN SMALL LETTER I ACUTE
                    case '\u00f3': replace = "\u00f2"; break; // LATIN SMALL LETTER O ACUTE
                    case '\u00fa': replace = "\u00f9"; break; // LATIN SMALL LETTER U ACUTE
                    case '\u0152': replace = "Oe"; break; // LATIN CAPITAL LETTER O E
                    case '\u0153': replace = "oe"; break; // LATIN SMALL LETTER O E
                    case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
                    case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
                    default: replace += c; break;
                    }
                }
		beginWord = Character.isWhitespace(c) ||
		    (c == '\u2019') || (c == '\'');
		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	else if (ruleset.equals("fr")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		if (! beginWord) c = Character.toLowerCase(c);
		String replace = new String();
	    
		if ((c == 'u') && (i > 0) && (i < s.length() - 1)) {
		    char c_n = s.charAt(i + 1);
		    char c_p;
		    int y = i - 1;
		    while ((((c_p = s.charAt(y)) == '-') ||
			    (c_p == '\u00ad')) && (y-- > 0));
		    // sic: OK, I think, to use IT_VOWELS
		    if ((IT_VOWELS.indexOf(c_p) > -1) &&
			(IT_VOWELS.indexOf(c_n) > -1)) replace = "v";
		    else replace = "u";
		}
		else if ((c == 'v') && (i < s.length() - 1)) {
		    char c_p;
		    char c_n;
		    int y = i + 1;
		    while ((((c_n = s.charAt(y)) == '-') ||
			    (c_n == '\u00ad')) && (y++ > 0));
		    if ((beginWord ||
			 (! Character.isLetter(c_p = s.charAt(i - 1))) ||
			 IT_CONS.indexOf(c_p) > -1) &&
			IT_CONS.indexOf(c_n) > -1)
			replace = "u"; else replace = "v";
		}
		else {
                    switch (c) {
                    case '\u00e6': replace = "ae"; break; // LATIN SMALL LETTER A E
                    case '\u017f': replace = "s"; break; // LATIN SMALL LETTER LONG S
                    case '\u00df': replace = "ss"; break; // LATIN SMALL LETTER SHARP S
                    case '\u00ad': break; // soft hyphen
                    case '-': break;
                    default: replace += c; break;
                    }
                }
		beginWord = Character.isWhitespace(c);
		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	else if (ruleset.equals("de")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		if (! beginWord) c = Character.toLowerCase(c);
		beginWord = Character.isWhitespace(c) ||
		    (" .,:;?!(<['\"\u2018\u201c".indexOf(c) > -1);
		String replace = new String();
	    
		switch (c) {
		case '\u00c4': replace = "Ae"; break;
		case '\u00d6': replace = "Oe"; break;
		case '\u00dc': replace = "Ue"; break;
		case '\u00df': replace = "ss"; break;
		case '\u00e4': replace = "ae"; break;
		case '\u00f6': replace = "oe"; break;
		case '\u00fc': replace = "ue"; break;
		case '\u00e9': replace = "e"; break;
		case '\u00ad': break; // soft hyphen
		case '-': break;
		default: replace += c; break;
		}

		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	else if (ruleset.equals("zh")) {
	    StringBuffer buf = new StringBuffer();
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		String replace = new String();
	    
		switch (c) {
		case '\u00b9': replace = "1"; break;
		case '\u00b2': replace = "2"; break;
		case '\u00b3': replace = "3"; break;
		case '\u2074': replace = "4"; break;
		case '\u2075': replace = "5"; break;
		case '\u3000': replace = " "; break;
		case '\u3001': replace = ","; break;
		case '\u3002': replace = "."; break;
// 		case '\u200b': break; // BREAKS EVERYTHING!
		default: replace += c; break;
		}

		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	else if (ruleset.equals("ar")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		beginWord = Character.isWhitespace(c);
		String replace = new String();
	    
		switch (c) {
		case '_': break; // TATWEEL in Buckwalter
		case 'o': break; // SUKUN in Buckwalter
		case '\u0640': break; // TATWEEL
		case '\u0652': break; // SUKUN
		case '^': break; // Buckwalter extension for capitalization
		case '\u200f': break; // R-L MARK
		default: replace += c; break;
		}

		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	else if (ruleset.equals("akk") ||
		 ruleset.equals("qam") ||
		 ruleset.equals("qpc") ||
		 ruleset.equals("elx") ||
		 ruleset.equals("sux") ||
		 ruleset.equals("hit") ||
		 ruleset.equals("qhu") ||
		 ruleset.equals("peo") ||
		 ruleset.equals("uga") ||
		 ruleset.equals("ura") ||
		 ruleset.equals("qcu")) {
	    

            StringBuffer buf = new StringBuffer();
            int n = 0;
	    char last = '\u0000';
            for (int i = 0; i < s.length(); i++) {
                char c = s.charAt(i);
                c = Character.toLowerCase(c);
                String replace = new String();

                switch (c) {
		case '{': replace += "-"; break;
		case '}': replace += "-"; break;

		// These are from PSD::ATF::Unicode by Steve Tinney
		case '\u0161': replace += "sz"; break;
		case '\u1e63': replace += "s,"; break;
		case '\u1e6d': replace += "t,"; break;
		case '\u014b': replace += "j"; break;
		case '\u015b': replace += "s'"; break;
		case '\u2080': replace += "0"; break;
		case '\u2081': replace += "1"; break;
		case '\u2082': replace += "2"; break;
		case '\u2083': replace += "3"; break;
		case '\u2084': replace += "4"; break;
		case '\u2085': replace += "5"; break;
		case '\u2086': replace += "6"; break;
		case '\u2087': replace += "7"; break;
		case '\u2088': replace += "8"; break;
		case '\u2089': replace += "9"; break;

		case 'c':	// shin (except where used as modifier)
		    if ((i > 0) && ((last == '~') || (last == '@')))
			replace += "c";
		    else replace += "sz";
		    break;

                default: replace += c; break;
                }

		// suppress grapheme boundary before or after word boundary
		if (replace.equals("-")) {
		    if ((i + 1 == s.length()) || (s.charAt(i + 1) == ' ') ||
			(i == 0) || (s.charAt(i - 1) == ' ')) replace = "";
		}

		last = c;

                buf.append(replace);

                // update offsets if replacement is a different length
                if (offsets != null) {
                    int r = replace.length();
                    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
                    else if (r == 2)
                        this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
                    n += 1 - r;
                }
            }
            return buf.toString();
	}

	if (ruleset.equals("el")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		if (! beginWord) c = Character.toLowerCase(c);
		beginWord = Character.isWhitespace(c);
		String replace = new String();

		switch (c) {
		case '\u03c2': replace = "\u03c3"; break; // GREEK SMALL LETTER FINAL SIGMA
		case '<': break;
		case '>': break;
		case '[': break;
		case ']': break;
		case '1': break;
		case '2': break;
		case '\u03ac': replace = "\u1f71"; break;
		case '\u03ad': replace = "\u1f73"; break;
		case '\u03ae': replace = "\u1f75"; break;
		case '\u03af': replace = "\u1f77"; break;
		case '\u03cc': replace = "\u1f79"; break;
		case '\u03cd': replace = "\u1f7b"; break;
		case '\u03ce': replace = "\uu1f7d"; break;
		case '-': break; // same treatment as soft hyphen
		case '\u00ad': break; // soft hyphen
		default: replace += c; break;
		}

		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	if (ruleset.equals("el_atonic")) {
	    StringBuffer buf = new StringBuffer();
	    boolean beginWord = true;
	    int n = 0;
	    for (int i = 0; i < s.length(); i++) {
		char c = s.charAt(i);
		if (! beginWord) c = Character.toLowerCase(c);
		beginWord = Character.isWhitespace(c);
		String replace = new String();

		switch (c) {
		case '\u03c2': replace = "\u03c3"; break; // GREEK SMALL LETTER FINAL SIGMA
		// map characters with diacritics to their plain equivalent
  	        // cf. <code>BetaCode.java</code>
		case '\u03aa': replace = "\u0399"; break;
		case '\u03ab': replace = "\u03a5"; break;
		case '\u03ac': replace = "\u0381"; break;
		case '\u03ad': replace = "\u0385"; break;
		case '\u03ae': replace = "\u0387"; break;
		case '\u03af': replace = "\u0389"; break;
		case '\u03ca': replace = "\u03b9"; break;
		case '\u03cb': replace = "\u03c5"; break;
		case '\u03cc': replace = "\u03bf"; break;
		case '\u03cd': replace = "\u03c5"; break;
		case '\u03ce': replace = "\u03c9"; break;
		case '\u1f00': replace = "\u03b1"; break;
		case '\u1f01': replace = "\u03b1"; break;
		case '\u1f02': replace = "\u03b1"; break;
		case '\u1f03': replace = "\u03b1"; break;
		case '\u1f04': replace = "\u03b1"; break;
		case '\u1f05': replace = "\u03b1"; break;
		case '\u1f06': replace = "\u03b1"; break;
		case '\u1f07': replace = "\u03b1"; break;
		case '\u1f08': replace = "\u0391"; break;
		case '\u1f09': replace = "\u0391"; break;
		case '\u1f0a': replace = "\u0391"; break;
		case '\u1f0b': replace = "\u0391"; break;
		case '\u1f0c': replace = "\u0391"; break;
		case '\u1f0d': replace = "\u0391"; break;
		case '\u1f0e': replace = "\u0391"; break;
		case '\u1f0f': replace = "\u0391"; break;
		case '\u1f10': replace = "\u03b5"; break;
		case '\u1f11': replace = "\u03b5"; break;
		case '\u1f12': replace = "\u03b5"; break;
		case '\u1f13': replace = "\u03b5"; break;
		case '\u1f14': replace = "\u03b5"; break;
		case '\u1f15': replace = "\u03b5"; break;
		case '\u1f18': replace = "\u0395"; break;
		case '\u1f19': replace = "\u0395"; break;
		case '\u1f1a': replace = "\u0395"; break;
		case '\u1f1b': replace = "\u0395"; break;
		case '\u1f1c': replace = "\u0395"; break;
		case '\u1f1d': replace = "\u0395"; break;
		case '\u1f20': replace = "\u03b7"; break;
		case '\u1f21': replace = "\u03b7"; break;
		case '\u1f22': replace = "\u03b7"; break;
		case '\u1f23': replace = "\u03b7"; break;
		case '\u1f24': replace = "\u03b7"; break;
		case '\u1f25': replace = "\u03b7"; break;
		case '\u1f26': replace = "\u03b7"; break;
		case '\u1f27': replace = "\u03b7"; break;
		case '\u1f28': replace = "\u0397"; break;
		case '\u1f29': replace = "\u0397"; break;
		case '\u1f2a': replace = "\u0397"; break;
		case '\u1f2b': replace = "\u0397"; break;
		case '\u1f2c': replace = "\u0397"; break;
		case '\u1f2d': replace = "\u0397"; break;
		case '\u1f2e': replace = "\u0397"; break;
		case '\u1f2f': replace = "\u0397"; break;
		case '\u1f30': replace = "\u03b9"; break;
		case '\u1f31': replace = "\u03b9"; break;
		case '\u1f32': replace = "\u03b9"; break;
		case '\u1f33': replace = "\u03b9"; break;
		case '\u1f34': replace = "\u03b9"; break;
		case '\u1f35': replace = "\u03b9"; break;
		case '\u1f36': replace = "\u03b9"; break;
		case '\u1f37': replace = "\u03b9"; break;
		case '\u1f38': replace = "\u0399"; break;
		case '\u1f39': replace = "\u0399"; break;
		case '\u1f3a': replace = "\u0399"; break;
		case '\u1f3b': replace = "\u0399"; break;
		case '\u1f3c': replace = "\u0399"; break;
		case '\u1f3d': replace = "\u0399"; break;
		case '\u1f3e': replace = "\u0399"; break;
		case '\u1f3f': replace = "\u0399"; break;
		case '\u1f40': replace = "\u03bf"; break;
		case '\u1f41': replace = "\u03bf"; break;
		case '\u1f42': replace = "\u03bf"; break;
		case '\u1f43': replace = "\u03bf"; break;
		case '\u1f44': replace = "\u03bf"; break;
		case '\u1f45': replace = "\u03bf"; break;
		case '\u1f48': replace = "\u039f"; break;
		case '\u1f49': replace = "\u039f"; break;
		case '\u1f4a': replace = "\u039f"; break;
		case '\u1f4b': replace = "\u039f"; break;
		case '\u1f4c': replace = "\u039f"; break;
		case '\u1f4d': replace = "\u039f"; break;
		case '\u1f50': replace = "\u03c5"; break;
		case '\u1f51': replace = "\u03c5"; break;
		case '\u1f52': replace = "\u03c5"; break;
		case '\u1f53': replace = "\u03c5"; break;
		case '\u1f54': replace = "\u03c5"; break;
		case '\u1f55': replace = "\u03c5"; break;
		case '\u1f56': replace = "\u03c5"; break;
		case '\u1f57': replace = "\u03c5"; break;
		case '\u1f58': replace = "\u03a5"; break;
		case '\u1f59': replace = "\u03a5"; break;
		case '\u1f5a': replace = "\u03a5"; break;
		case '\u1f5b': replace = "\u03a5"; break;
		case '\u1f5c': replace = "\u03a5"; break;
		case '\u1f5d': replace = "\u03a5"; break;
		case '\u1f5e': replace = "\u03a5"; break;
		case '\u1f5f': replace = "\u03a5"; break;
		case '\u1f60': replace = "\u03c9"; break;
		case '\u1f61': replace = "\u03c9"; break;
		case '\u1f62': replace = "\u03c9"; break;
		case '\u1f63': replace = "\u03c9"; break;
		case '\u1f64': replace = "\u03c9"; break;
		case '\u1f65': replace = "\u03c9"; break;
		case '\u1f66': replace = "\u03c9"; break;
		case '\u1f67': replace = "\u03c9"; break;
		case '\u1f68': replace = "\u03a9"; break;
		case '\u1f69': replace = "\u03a9"; break;
		case '\u1f6a': replace = "\u03a9"; break;
		case '\u1f6b': replace = "\u03a9"; break;
		case '\u1f6c': replace = "\u03a9"; break;
		case '\u1f6d': replace = "\u03a9"; break;
		case '\u1f6e': replace = "\u03a9"; break;
		case '\u1f6f': replace = "\u03a9"; break;
		case '\u1f70': replace = "\u03b1"; break;
		case '\u1f71': replace = "\u03b1"; break;
		case '\u1f72': replace = "\u03b5"; break;
		case '\u1f73': replace = "\u03b5"; break;
		case '\u1f74': replace = "\u03b7"; break;
		case '\u1f75': replace = "\u03b7"; break;
		case '\u1f76': replace = "\u03b9"; break;
		case '\u1f77': replace = "\u03b9"; break;
		case '\u1f78': replace = "\u03bf"; break;
		case '\u1f79': replace = "\u03bf"; break;
		case '\u1f7a': replace = "\u03c5"; break;
		case '\u1f7b': replace = "\u03c5"; break;
		case '\u1f7c': replace = "\u03c9"; break;
		case '\u1f7d': replace = "\u03c9"; break;
		case '\u1f80': replace = "\u03b1"; break;
		case '\u1f81': replace = "\u03b1"; break;
		case '\u1f82': replace = "\u03b1"; break;
		case '\u1f83': replace = "\u03b1"; break;
		case '\u1f84': replace = "\u03b1"; break;
		case '\u1f85': replace = "\u03b1"; break;
		case '\u1f86': replace = "\u03b1"; break;
		case '\u1f87': replace = "\u03b1"; break;
		case '\u1f88': replace = "\u0391"; break;
		case '\u1f89': replace = "\u0391"; break;
		case '\u1f8a': replace = "\u0391"; break;
		case '\u1f8b': replace = "\u0391"; break;
		case '\u1f8c': replace = "\u0391"; break;
		case '\u1f8d': replace = "\u0391"; break;
		case '\u1f8e': replace = "\u0391"; break;
		case '\u1f8f': replace = "\u0391"; break;
		case '\u1f90': replace = "\u03b7"; break;
		case '\u1f91': replace = "\u03b7"; break;
		case '\u1f92': replace = "\u03b7"; break;
		case '\u1f93': replace = "\u03b7"; break;
		case '\u1f94': replace = "\u03b7"; break;
		case '\u1f95': replace = "\u03b7"; break;
		case '\u1f96': replace = "\u03b7"; break;
		case '\u1f97': replace = "\u03b7"; break;
		case '\u1f98': replace = "\u0397"; break;
		case '\u1f99': replace = "\u0397"; break;
		case '\u1f9a': replace = "\u0397"; break;
		case '\u1f9b': replace = "\u0397"; break;
		case '\u1f9c': replace = "\u0397"; break;
		case '\u1f9d': replace = "\u0397"; break;
		case '\u1f9e': replace = "\u0397"; break;
		case '\u1f9f': replace = "\u0397"; break;
		case '\u1fa0': replace = "\u03c9"; break;
		case '\u1fa1': replace = "\u03c9"; break;
		case '\u1fa2': replace = "\u03c9"; break;
		case '\u1fa3': replace = "\u03c9"; break;
		case '\u1fa4': replace = "\u03c9"; break;
		case '\u1fa5': replace = "\u03c9"; break;
		case '\u1fa6': replace = "\u03c9"; break;
		case '\u1fa7': replace = "\u03c9"; break;
		case '\u1fa8': replace = "\u03a9"; break;
		case '\u1fa9': replace = "\u03a9"; break;
		case '\u1faa': replace = "\u03a9"; break;
		case '\u1fab': replace = "\u03a9"; break;
		case '\u1fac': replace = "\u03a9"; break;
		case '\u1fad': replace = "\u03a9"; break;
		case '\u1fae': replace = "\u03a9"; break;
		case '\u1faf': replace = "\u03a9"; break;
		case '\u1fb2': replace = "\u03b1"; break;
		case '\u1fb3': replace = "\u03b1"; break;
		case '\u1fb4': replace = "\u03b1"; break;
		case '\u1fb6': replace = "\u03b1"; break;
		case '\u1fb7': replace = "\u03b1"; break;
		case '\u1fba': replace = "\u0391"; break;
		case '\u1fbb': replace = "\u0391"; break;
		case '\u1fbc': replace = "\u0391"; break;
		case '\u1fc2': replace = "\u03b7"; break;
		case '\u1fc3': replace = "\u03b7"; break;
		case '\u1fc4': replace = "\u03b7"; break;
		case '\u1fc6': replace = "\u03b7"; break;
		case '\u1fc7': replace = "\u03b7"; break;
		case '\u1fca': replace = "\u0397"; break;
		case '\u1fcb': replace = "\u0397"; break;
		case '\u1fcc': replace = "\u0397"; break;
		case '\u1fd2': replace = "\u03b9"; break;
		case '\u1fd3': replace = "\u03b9"; break;
		case '\u1fd6': replace = "\u03b9"; break;
		case '\u1fd7': replace = "\u03b9"; break;
		case '\u1fda': replace = "\u0399"; break;
		case '\u1fdb': replace = "\u039f"; break;
		case '\u1fe2': replace = "\u03c5"; break;
		case '\u1fe3': replace = "\u03c5"; break;
		case '\u1fe4': replace = "\u03c1"; break;
		case '\u1fe5': replace = "\u03c1"; break;
		case '\u1fe6': replace = "\u03c5"; break;
		case '\u1fe7': replace = "\u03c5"; break;
		case '\u1fea': replace = "\u03a5"; break;
		case '\u1feb': replace = "\u03a5"; break;
		case '\u1fec': replace = "\u03a1"; break;
		case '\u1ff2': replace = "\u03c9"; break;
		case '\u1ff3': replace = "\u03c9"; break;
		case '\u1ff4': replace = "\u03c9"; break;
		case '\u1ff6': replace = "\u03c9"; break;
		case '\u1ff7': replace = "\u03c9"; break;
		case '\u1ff8': replace = "\u039f"; break;
		case '\u1ff9': replace = "\u039f"; break;
		case '\u1ffa': replace = "\u03a9"; break;
		case '\u1ffb': replace = "\u03a9"; break;
		case '\u1ffc': replace = "\u03a9"; break;

		case '<': break;
		case '>': break;
		case '-': break; // same treatment as soft hyphen
		case '\u00ad': break; // soft hyphen
		default: replace += c; break;
		}

		buf.append(replace);

		// update offsets if replacement is a different length
		if (offsets != null) {
		    int r = replace.length();
		    if (r == 0) this.offsets = arrayKill(this.offsets, i - n);
		    else if (r == 2)
			this.offsets =
			    arrayInsert(this.offsets, i - n + 1,
					this.offsets[i - n], r - 1);
		    n += 1 - r;
		}
	    }
	    return buf.toString();
	}

	else {			// unknown or no ruleset
	    return s;
	}
    }

    /**
     * Applies the normalization rules in <code>ruleset</code> to
     * <code>s</code>, without offset tracking.
     *
     * @param s            source string
     * @return             normalized string
     */
    public String normalize(String s) {
	return normalize(s, null);
    }

    /**
     * Returns the offset table.
     *
     * @return             offset table
     */
    public int[] getOffsetTable() {
	return offsets;
    }

    /**
     * Returns a copy of an integer array with the element at
     * <code>index</code> removed ("killed").
     *
     * @param array        integer array
     * @param index        index of element to remove
     */
    static private int[] arrayKill(int[] array, int index) {
	int[] newArray = new int[array.length - 1];
	System.arraycopy(array, 0, newArray, 0, index);
	System.arraycopy(array, index + 1, newArray, index,
			 array.length - index - 1);
	return newArray;
    }

    /**
     * Returns a copy of an integer array with <code>count</code> elements
     * inserted at <code>index</code>.
     *
     * @param array        integer array
     * @param index        index to insert new elements
     * @param value        value to insert into new slots
     * @param count        number of new slots to insert
     */
    static private int[] arrayInsert(int[] array, int index, int value,
				     int count) {
	int[] newArray = new int[array.length + count];
	System.arraycopy(array, 0, newArray, 0, index);
	for (int i = 0; i < count; i++) newArray[index + i] = value;
	System.arraycopy(array, index, newArray, index + count,
			 array.length - index);
	return newArray;
    }

    /**
     * We provide <code>main()</code> so that our services will be available
     * outside Java (i.e., so we can run as a Un*x-style filter).
     */
    static public void main(String[] argv) {
	if (argv.length != 1) {
	    System.err.println("You must specify a ruleset.");
	    System.exit(1);
	}

	String rec;
	StringBuffer buf = new StringBuffer();
	BufferedReader bin = null;
	try {
	    bin = new BufferedReader(new InputStreamReader(System.in,
							   "UTF-8"));
	    while ((rec = bin.readLine()) != null) buf.append(rec + "\n");
	}
	catch (UnsupportedEncodingException e) {
	    System.err.println(e);
	    System.exit(1);
	}
	catch (IOException e) {
	    System.err.println(e);
	    System.exit(1);
	}

	OrthographicRules orth = new OrthographicRules(argv[0]);
	System.out.print(orth.normalize(buf.toString()));
    }
}
